unit unitMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, cbAudioPlay, ExtCtrls, Menus, DirectShow, ComCtrls, Buttons,
  MMSystem;

type
  TForm1 = class(TForm)
    btnPlay: TButton;
    btnStop: TButton;
    btnPause: TButton;
    Timer1: TTimer;
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    mnuDevice: TMenuItem;
    Open1: TMenuItem;
    Close1: TMenuItem;
    N1: TMenuItem;
    Exit1: TMenuItem;
    mnuDefault: TMenuItem;
    OpenDialog1: TOpenDialog;
    lblDuration: TLabel;
    GroupBox1: TGroupBox;
    ckbSeekAbsolute: TCheckBox;
    ckbSeekForwards: TCheckBox;
    ckbSeekBackwards: TCheckBox;
    ckbGetCurrentPos: TCheckBox;
    ckbGetStopPos: TCheckBox;
    ckbGetDuration: TCheckBox;
    ckbPlayBackward: TCheckBox;
    ckbDoSeamlessLooping: TCheckBox;
    tbarPosition: TTrackBar;
    lblPosition: TLabel;
    Label1: TLabel;
    tbarBalance: TTrackBar;
    Label2: TLabel;
    Label3: TLabel;
    tbarVolume: TTrackBar;
    sbtnLoop: TSpeedButton;
    procedure FormShow(Sender: TObject);
    procedure btnPlayClick(Sender: TObject);
    procedure btnStopClick(Sender: TObject);
    procedure btnPauseClick(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure Exit1Click(Sender: TObject);
    procedure Open1Click(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure mnuDefaultClick(Sender: TObject);
    procedure Close1Click(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure tbarPositionChange(Sender: TObject);
    procedure tbarBalanceChange(Sender: TObject);
    procedure tbarVolumeChange(Sender: TObject);
    procedure sbtnLoopClick(Sender: TObject);
  private
    FDSAP: TcbAudioPlay;
    FDeviceIdx: Integer;
    FIsMediaTime,
    FDontSeek,
    FDontBalance,
    FDontVolume,
    FCanSeek: Boolean;

    function LongTimeToStr(Time: LONGLONG): String;
    procedure OpenAudioPlay(FileName: String);
    procedure FreeAudioPlay;
    procedure RefreshCurrentPosition;
    procedure RefreshFileInfo;
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.DFM}

const
  BAR_MAX = 100;
  BAR_MIN = 0;
  RATE_MAX = 250;
  RATE_MIN = 0;
  RATE_NORMAL = 100;
  APP_TITLE = ' - cbAudioPlay Example (www.carlosb.com)';
  APP_INIFILE = 'example.ini';

procedure TForm1.FormShow(Sender: TObject);
var
  Cnt: Integer;
  MenuItem: TMenuItem;
begin
  for Cnt := 0 to cbDeviceList.count-1 do
  begin
    MenuItem := TMenuItem.Create(Self);
    MenuItem.Caption := cbDeviceList[cnt];
    MenuItem.Tag := Cnt;
    MenuItem.GroupIndex := 1;
    MenuItem.RadioItem := True;
    MenuItem.OnClick := mnuDefaultClick;

    mnuDevice.Add(MenuItem);
  end;

  RefreshFileInfo;
end;

procedure TForm1.btnPlayClick(Sender: TObject);
begin
  FDSAP.Play;
end;

procedure TForm1.btnStopClick(Sender: TObject);
begin
  FDSAP.Stop;
end;

procedure TForm1.btnPauseClick(Sender: TObject);
begin
  FDSAP.Pause;
end;

procedure TForm1.Timer1Timer(Sender: TObject);
begin
  if Assigned(FDSAP) and (FDSAP.Duration <> 0) then
    RefreshCurrentPosition;
end;

procedure TForm1.Exit1Click(Sender: TObject);
begin
  Close;
end;

procedure TForm1.Open1Click(Sender: TObject);
begin
  if OpenDialog1.Execute then
  begin
    FreeAudioPlay;
    OpenAudioPlay(OpenDialog1.FileName);
    RefreshFileInfo;
  end;
end;

procedure TForm1.OpenAudioPlay(FileName: String);
var
  tf: TGUID;
begin
  if Assigned(FDSAP) then
    FreeAudioPlay;

  FDSAP := TcbAudioPlay.Create(self, FileName, FDeviceIdx);
  FDSAP.GetTimeFormat(tf);
  if CompareMem(@tf, @TIME_FORMAT_MEDIA_TIME, sizeof(TGUID)) then
    FIsMediaTime := True
  else
    FIsMediaTime := False;
  FCanSeek := (FDSAP.Capabilities and AM_SEEKING_CanSeekAbsolute) <> 0;
end;

procedure TForm1.FreeAudioPlay;
begin
  if Assigned(FDSAP) then
  begin
    FDSAP.Free;
    FDSAP := nil;
  end;
end;

procedure TForm1.RefreshFileInfo;
var
  Capab,
  Balance,
  Volume: Longint;
begin
  if Assigned(FDSAP) then
  begin
    Caption := FDSAP.FileName + APP_TITLE;
    btnPlay.Enabled := True;
    btnStop.Enabled := True;
    btnPause.Enabled := True;
    tbarPosition.Enabled := True;
    tbarBalance.Enabled := True;
    tbarVolume.Enabled := True;
    sbtnLoop.Enabled := True;
    Capab := FDSAP.Capabilities;
    lblDuration.Caption := LongTimeToStr(FDSAP.Duration);
    Balance := FDSAP.Balance;
    Volume := FDSAP.Volume;
    sbtnLoop.Down := FDSAP.Loop;
  end
  else
  begin
    Caption := '(none)' + APP_TITLE;
    btnPlay.Enabled := False;
    btnStop.Enabled := False;
    btnPause.Enabled := False;
    tbarPosition.Enabled := False;
    tbarBalance.Enabled := False;
    tbarVolume.Enabled := False;
    sbtnLoop.Enabled := False;
    Capab := 0;
    lblDuration.Caption := LongTimeToStr(0);
    Balance := (cbAudioPlay_MAX_BALANCE + cbAudioPlay_MIN_BALANCE) div 2;
    Volume := cbAudioPlay_MAX_VOLUME;
    sbtnLoop.Down := False;
  end;

  ckbSeekAbsolute.Checked := (Capab and AM_SEEKING_CanSeekAbsolute) <> 0;
  ckbSeekForwards.Checked := (Capab and AM_SEEKING_CanSeekForwards) <> 0;
  ckbSeekBackwards.Checked := (Capab and AM_SEEKING_CanSeekBackwards) <> 0;
  ckbGetCurrentPos.Checked := (Capab and AM_SEEKING_CanGetCurrentPos) <> 0;
  ckbGetStopPos.Checked := (Capab and AM_SEEKING_CanGetStopPos) <> 0;
  ckbGetDuration.Checked := (Capab and AM_SEEKING_CanGetDuration) <> 0;
  ckbPlayBackward.Checked := (Capab and AM_SEEKING_CanPlayBackwards) <> 0;
  ckbDoSeamlessLooping.Checked := (Capab and AM_SEEKING_CanDoSegments) <> 0;

  FDontBalance := True;
  tbarBalance.Position := Balance;
  FDontBalance := False;
  FDontVolume := True;
  tbarVolume.Position := Volume;
  FDontVolume := False;

  RefreshCurrentPosition;
end;

procedure TForm1.RefreshCurrentPosition;
var
  OldDontSeek: Boolean;
begin
  OldDontSeek := FDontSeek;
  FDontSeek := True;
  try
    if Assigned(FDSAP) then
    begin
      if (FDSAP.Capabilities and AM_SEEKING_CanGetDuration) <> 0 then
        tbarPosition.Position := Trunc((FDSAP.CurrentPosition*BAR_MAX) /
          FDSAP.Duration);
      lblPosition.Caption := LongTimeToStr(FDSAP.CurrentPosition)
    end
    else
    begin
      tbarPosition.Position := 0;
      lblPosition.Caption := LongTimeToStr(0);
    end;
  finally
    FDontSeek := OldDontSeek;
  end;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  FreeAudioPlay;
end;

procedure TForm1.mnuDefaultClick(Sender: TObject);
begin
  if not TMenuItem(Sender).Checked then
  begin
    TMenuItem(Sender).Checked := True;
    FDeviceIdx := TMenuItem(Sender).Tag;

    if Assigned(FDSAP) then
      ShowMessage('The selected device will be used only the next time you open a file');
  end;
end;

procedure TForm1.Close1Click(Sender: TObject);
begin
  FreeAudioPlay;
  RefreshFileInfo;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  tbarPosition.Min := BAR_MIN;
  tbarPosition.Max := BAR_MAX;
  FIsMediaTime := True;
  tbarBalance.Min := cbAudioPlay_MIN_BALANCE;
  tbarBalance.Max := cbAudioPlay_MAX_BALANCE;
  tbarVolume.Min := cbAudioPlay_MIN_VOLUME;
  tbarVolume.Max := cbAudioPlay_MAX_VOLUME;
end;

procedure TForm1.tbarPositionChange(Sender: TObject);
var
  Pos: LONGLONG;
begin
  if not FDontSeek and FCanSeek then
  begin
    Pos := Trunc((tbarPosition.Position * FDSAP.Duration) / 100);
    FDSAP.SetCurrentPosition(Pos, AM_SEEKING_AbsolutePositioning);
  end;

  RefreshCurrentPosition;
end;

function TForm1.LongTimeToStr(Time: LONGLONG): String;
begin
  if FIsMediaTime then
    Result := FormatDateTime('hh:nn:ss:zzz',
      Time / (10000000.0*24.0*60.0*60.0))
  else
    Result := IntToStr(Time);
end;

procedure TForm1.tbarBalanceChange(Sender: TObject);
begin
  if Assigned(FDSAP) and (not FDontBalance) then
    FDSAP.Balance := tbarBalance.Position;
end;

procedure TForm1.tbarVolumeChange(Sender: TObject);
begin
  if Assigned(FDSAP) and (not FDontVolume) then
    FDSAP.Volume := tbarVolume.Position;
end;

procedure TForm1.sbtnLoopClick(Sender: TObject);
begin
  if Assigned(FDSAP) then
    FDSAP.Loop := sbtnLoop.Down;
end;

end.
